"""MOSFiT: Modular light curve fitting software."""
from . import constants  # noqa: F401
from . import fitter  # noqa: F401
from . import model  # noqa: F401
from . import plotting  # noqa: F401
from . import printer  # noqa: F401
from . import utils  # noqa: F401

__version__ = '0.5.11'
__author__ = 'James Guillochon & Matt Nicholl'
__license__ = 'MIT'
